
#define VERSION "ryb930803"

/*

  smodemdl - Downloads DSP code to a DSI Connection 96+ modem under Linux.
  Copyright (C) 1993 Ray Burr

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  ----------------------------------------------------------------------

  This is a reverse-engineered version of Digicom's `dl.exe'.  I do
  not have any real documentation on how to download the modem.  All of
  this code is based on observations of the port I/O that `dl.exe' makes
  under dosemu (thanks Robert Sanders, Matthias Lautner, and others!)
  and a very little peeking at the disassembled code.  I have never seen
  any of the original source code.

  IMPORTANT: This program is not supported by Digicom.  I am nothing
  more than a customer of Digicom.  If you have problems, please be sure
  that it is not due to this program before you bother their tech
  support people about it.  If you have a problem with this program,
  send me e-mail.

  To use this program, you must run it as root and give it a filename
  and a com port number (1 to 4).  It will download the modem or print
  an error message if there was a problem.  You might want to invoke it
  in one of your boot scripts (/etc/rc.local or equivalent).

  Example:

    smodemdl /etc/smodem/data114 1

  This would download a modem on ttyS0 (COM1) with the file
  /etc/smodem/data114.

  Ray Burr  <ryb@hoggle.oau.org>

*/


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>


#define PORT_BASE_DEFAULT 0x2e8	/* COM4 */

#define PORT_LEN 8

#define SANE_FILE_SIZE 200000

#define RAM_DATA 0
#define RAM_CODE 1

/* [Adapted from /usr/include/linux/serial.h (C) 1992 by Theodore Ts'o.]
 * These are the UART port assignments, expressed as offsets from the base
 * register.
 */
#define UART_DLL	0	/* Out: Devisor Latch Low (DLAB=1) */
#define UART_DLM	1	/* Out: Devisor Latch High (DLAB=1) */
#define UART_LCR	3	/* Out: Line Control Register */
#define UART_MCR	4	/* Out: Modem Control Register */

#define UART_LCR_DLAB	0x80	/* Devisor latch access bit */

#define UART_MCR_OUT2	0x08	/* Out2 complement */
#define UART_MCR_OUT1	0x04	/* Out1 complement */


unsigned long port_base;	/* The base address of the card. */


/* In and out to port at PORT_BASE + OFFSET. */

inline static void port_out (unsigned short offset, char value)
{
__asm__ volatile ("outb %0,%1"
		  : :"a" ((char) value),"d" ((unsigned short)
					     (port_base + offset)));
}

inline static char port_in (unsigned short offset)
{
	char _v;
__asm__ volatile ("inb %1,%0"
		  :"=a" (_v):"d" ((unsigned short) (port_base + offset)));
	return _v;
}


/* Select between data (TYPE == 0) and program (TYPE == 1) RAM. */

int
select_ram_type (int type)
{
  static int current_type = -1;

  if (current_type < 0)
    current_type = port_in (7) & 1;
  if (current_type != type)
    {
      port_out (7, 0x18);
      port_out (7, 0x19);
    }
  current_type = port_in (7) & 1;
  return (current_type != type);
}


/* Optionally print an error message, then exit with error status. */

void
die (const char *s)
{
  if (s)
    {
      fprintf (stderr, "%s", s);
      fputc ('\n', stderr);
    }
  exit (1);
}


/* Like die(), but for I/O errors. */

void
pdie (const char *s)
{
  perror (s);
  exit (1);
}


/* This is it. */

void
main (int argc, char *argv[])
{
  int i;
  int file_len;
  FILE *fp;
  char *file;
  unsigned char *buf, *p;
  int old_lcr, old_mcr;

  if (argc < 2 || argc > 3)
    {
      printf ("smodemdl - Downloads DSP code to a DSI Connection 96+ modem\n");
      printf ("Version: " VERSION "\n");
      printf ("Usage: dsidl file [port]\n");
      printf ("  port is 1 to 4 corresponding to MSDOS COM numbers\n");
      exit (1);
    }

  if (argc > 2)
    {
      int port_num = atoi (argv[2]);
      if (port_num < 1 || port_num > 4)
	{
	  printf ("The port number must be 1, 2, 3, or 4.\n");
	  exit (1);
	}
      port_base = ((int []){ 0x3f8, 0x2f8, 0x3e8, 0x2e8 })[port_num - 1];
    }
  else
    port_base = PORT_BASE_DEFAULT;

  file = argv[1];
  fp = fopen (file, "r");
  if (fp == NULL)
    pdie (file);

  if (fread (&file_len, 4, 1, fp) != 1)
    pdie (file);

  if (file_len > SANE_FILE_SIZE)
    die ("file size is bad");

  buf = malloc (file_len);
  if (buf == NULL)
    die ("malloc failed");

  if (fread (buf, 1, file_len, fp) != file_len)
    pdie (file);

  fclose (fp);

  p = buf;

  if (ioperm (port_base, PORT_LEN, 1))
    die ("couldn't get i/o perms");

  /* Bit 7 of port at offset 7 toggles each time is read. */

  if (((port_in (7) ^ port_in (7)) & 0x80) == 0)
    die ("can't talk to the card");

  old_mcr = port_in (UART_MCR);
  old_lcr = port_in (UART_LCR);

  /* Turning "OUT1" on does something. */

  port_out (UART_LCR, UART_LCR_DLAB);
  port_out (UART_MCR, UART_MCR_OUT1 | UART_MCR_OUT2);

  /* This seems to be a magic number. */

  port_out (UART_DLL, 0x31);
  port_out (UART_DLM, 0x8c);

  /* At this point, the card is in some sort of programming mode.
     The port at offset 7 is some sort of control/status register,
     and ports at 0, 1, and 2 are where the bytes are sent. */

  port_out (7, 0x10);
  port_out (7, 0x19);

  if ((port_in (7) & 0x1a) != 0x18)
    die ("DSP bus not granted");

  if (select_ram_type (RAM_CODE))
    die ("could not select program bus");

  for (i = 0; i < 7 * 0x800; i++)
    {
      port_out (0, *p++);
      port_out (1, *p++);
      port_out (2, *p++);
    }

  if (select_ram_type (RAM_DATA))
    die ("could not select data bus");
  
  for (i = 0; i < 2047; i++)
    {
      port_out (1, *p++);
      port_out (2, 0   );
      port_out (1, *p++);
      port_out (2, *p++);
    }
  port_out (1, 0);
  port_out (2, 0);
  p++;

  for (i = 0; i < 9945; i++)
    {
      port_out (1, *p++);
      port_out (2, *p++);
    }

  /* Make it go. */

  port_out (7, 0x11);
  port_out (7, 0x03);

  port_out (UART_MCR, old_mcr & ~UART_MCR_OUT1);
  port_out (UART_LCR, old_lcr);

  if (ioperm (port_base, PORT_LEN, 0))
    die ("couldn't clear i/o perms");

  exit (0);
}
